"""
The **asdf** submodule contains code that is used to serialize specutils types
so that they can be represented and stored using the Advanced Scientific Data
Format (ASDF).

If both **asdf** and **specutils** are installed, no further configuration is
required in order to process ASDF files that contain **specutils** types. The
**asdf** package has been designed to automatically detect the presence of the
tags defined by **specutils**.

Documentation on the ASDF Standard can be found `here
<https://asdf-standard.readthedocs.io>`__. Documentation on the ASDF Python
module can be found `here <https://asdf.readthedocs.io>`__. Additional details
for specutils developers can be found in :ref:`asdf_dev`.
"""
